function [model] = nuc_superhmm_v5_with_blast(hyb, xhyb, ent, varargin)
%date 1/12/05: modified from version 4. clear nucs may be 6-8 probes long.
% This is a modified version of Jong's original. 
% 1. nucs can be clear (7-8 probes long) or fuzzy (>9 probes long)
%    mean_n and mean_l is the same for clear and fuzzy nucs
% 2. The Blast node is eliminated from the HMM model. Instead, the hyb data are
% globally adjusted for blast-entropy effects prior to HMM in order to obtain reliable estimates 
% for mu_N and mu_L
%
%(original note by Jong).
%
% NUC_SUPERHMM
%
%       H_1 --> H_2 --> ... --> H_N
%        |       |               |
%        V       V               V
%       H'1     H'2     ...     H'N
%        |       |               |
%        V       V               V
%       O_1     O_2     ...     O_N
%
%    Hidden state transition diagram:
%
%       .-> n3 -> n4 -> n5 -> n6 -.
%       |                         |
%       n2                        n7
%       ^                         |
%       |                         V
%       n1 <--------------------- l8
%                                 ^ \
%                                 | |
%                                 `-'
%
%       n1 ... n7 model probes in a nucleosome.
%       l8 models probes in a linker.
%
%            ( 1,...,7   for nucleosome
%      H_i = -
%            (    8      for linker
%
%            ( 1 for nucleosome
%      H'i = -
%            ( 2 for linker
%
%      O_i = microarray data
%
%    Not shown here but each O_i also has a continuous parent B_i with two
%    observed symbols:
%       - BLAST score
%       - Entropy
%
%    Required arguments:
%       hyb	RxN matrix of R replicates, each with N observations
%	xhyb	1xN vector of BLAST scores
%       ent	1xN vector of entropy values
%
%    Optional arguments [with default values]:
%       PriorP		prior probability of staying in a linker [0.99]
%
% NOTE: Had to modify FullBNT/KPMstats/gaussian_prob.m.
% NOTE: Hyb data do not get inverted in this script.
% NOTE: This differs from v2 in that since Guocheng has done a great job with
%       normalizing across replicates, I am learning only one set of means,
%       standard deviations, and weights for all the replicates, whereas before,
%       I was learning one set for each replicate.

%P = 0.99;

tie_para = 0;  % Tying the parameters for different replicates together
tie_fuz  = 1;  % Tying the mean/var of fuzzy nuc with clear nuc
argv = varargin;
for i=1:2:length(argv)
    switch argv{i},
        case 'TiePara',     tie_para = argv{i+1};
        case 'TieFuz',      tie_fuz  = argv{i+1};
        otherwise,		error(['invalid argument name ' argv{i}]);
    end
end


[R LEN] = size(hyb);

% Slice size
ss = 3 + R;

% Topology within a slice, with nodes in this order:
%    H, H', B, O^1, O^2,...
intra = zeros(ss);

% H --> H'
intra(1,2) = 1;

% H',B --> O
intra([2 3],[4:ss]) = 1;

% Topology between slices
inter = zeros(ss);

% H_i --> H_(i+1)
inter(1,1) = 1;

% Number of states and observable symbols in this order:
%    H, H', B, O^1, O^2,...
states = zeros(1, ss);
states(1) = 18;
states(2) = 3;
states(3) = 2;
states(4:ss) = 1;

dnodes = [1 2];		% Discrete nodes
onodes = [3:ss];	% Observed nodes
eclass1 = [1:ss];	% Equivalence classes for the first slice
if tie_para == 1,
    eclass1(4:end) = 4;
end
eclass2 = eclass1;	% Equivalence classes for all other slices
eclass2(1) = eclass1(end)+1;

% This should actually take into account the abundance of nucleosomes on
% the yeast genome, rather than being a dumb prior.
prior0 = [0.01*ones(1, 17) 0.83];

PSNL = 0.8;
PNL = 0.8;
PFL = 0.8;
PLN = 0.1;
PLL = 0.8;

% State transition table for H.
% N1, ..., N8, F1, ..., F9, L
transit0 = zeros(18, 18);
for i = 1:5,
    transit0(i, i+1) = 1;
end
transit0(6, 7) = 1 - PSNL;
transit0(6, 18) = PSNL;
transit0(7, 8) = 1 - PNL;
transit0(7, 18) = PNL;
transit0(8, 18) = 1;
for i = 9:16,
    transit0(i, i+1) = 1;
end
transit0(17, 17) = 1 - PFL;
transit0(17, 18) = PFL;
transit0(18, 1) = PLN;
transit0(18, 9) = 1 - PLN - PLL;
transit0(18, 18) = PLL;


% A deterministic tying together of H states into three H' states.
if ~tie_fuz,
    merge0 = [ 1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        0 1 0;...
        0 1 0;...
        0 1 0;...
        0 1 0;...
        0 1 0;...
        0 1 0;...
        0 1 0;...
        0 1 0;...
        0 1 0;...
        0 0 1];
else,
    merge0 = [ 1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        1 0 0;...
        0 0 1];
end
% Create the DBN.
bnet = mk_dbn(intra, inter, states, 'discrete', dnodes,...
    'eclass1', eclass1, 'eclass2', eclass2, 'observed', onodes);

% Initialize the CPDs.
bnet.CPD{1} = tabular_CPD(bnet, bnet.rep_of_eclass(1), 'CPT', prior0);
bnet.CPD{2} = tabular_CPD(bnet, bnet.rep_of_eclass(2), 'CPT', merge0,...
    'adjustable', 0);
bnet.CPD{3} = root_CPD(bnet, bnet.rep_of_eclass(3), [0; 0]);
bnet.CPD{eclass2(1)} = tabular_CPD(bnet, bnet.rep_of_eclass(eclass2(1)), 'CPT', transit0);


if tie_para == 1,
    % Estimate priors for the observed nodes.
    m = median(hyb(:));
    above = hyb(hyb(:) >= m);
    below = hyb(hyb(:) < m);

    % Mean and variance of the microarray data for each H' state.
    % mean0 = zeros(symbols in O, states in H');
    mean0 = zeros(1,3);
    mean0 = [prctile(above, 90) prctile(below, 10), prctile(above, 90)];
    % var0 = zeros(symbols in O, symbols in O, states in H');
    var0 = ones(1,1,3);
    var0(:,:,:) = reshape([var(above) var(below) var(above)], 1, 1, 3);

    % The weight of B on O.
    % weight0 = zeros(symbols in O, symbls in B, states in H');
    weight0 = zeros(1,2,3) + 0.01;

    % Observed nodes.
    bnet.CPD{4} = gaussian_CPD(bnet, bnet.rep_of_eclass(4),...
        'mean', mean0, 'cov', var0, 'weights', weight0, 'cov_prior_weight', 0);

    engine = smoother_engine(jtree_2TBN_inf_engine(bnet));

    % Gather evidence.
    cases = cell(1,1);
    cases{1} = cell(ss, LEN);
    cases{1}(3,:) = num2cell([(xhyb > 0); ent], 1);
    cases{1}([4:ss],:) = num2cell(hyb);

    fprintf('Learning..
    mode.\n');
    [bnet, LLtrace, engine] = learn_params_dbn_em(engine, cases);

    fprintf('Tabulating...\n');

    % Store the learned parameters.
    s = struct(bnet.CPD{5});
    model.psnl = s.CPT(6, 18);
    model.pnl = s.CPT(7, 18);
    model.pfl = s.CPT(17, 18);
    model.pln = s.CPT(18, 1);
    model.pll = s.CPT(18, 18);
    s = struct(bnet.CPD{4});
    model.mean_n = s.mean(1,1);
    model.mean_f = s.mean(1,2);
    model.mean_l = s.mean(1,3);
    model.var_n = s.cov(1,1,1);
    model.var_f = s.cov(1,1,2);
    model.var_l = s.cov(1,1,3);
    model.blast_weight_n = s.weights(1,1,1);
    model.blast_weight_f = s.weights(1,1,2);
    model.blast_weight_l = s.weights(1,1,3);
    model.entropy_weight_n = s.weights(1,2,1);
    model.entropy_weight_f = s.weights(1,2,2);
    model.entropy_weight_l = s.weights(1,2,3);
    
    if tie_fuz,
        model.mean_f = model.mean_n;
        model.var_f = model.var_n;
        model.blast_weight_f = model.blast_weight_n;
        model.entropy_weight_f = model.entropy_weight_n;
    end
        
else,
    for i = 1:R
        % Estimate priors for the observed nodes.
        m = mean(hyb(i,:));
        above = hyb(i,hyb(i,:) >= m);
        below = hyb(i,hyb(i,:) < m);

        % Mean and variance of the microarray data for each H' state.
        % mean0 = zeros(symbols in O, states in H');
        mean0 = zeros(1,3);
        mean0 = [mean(above) mean(below) mean(above)];
        % var0 = zeros(symbols in O, symbols in O, states in H');
        var0 = ones(1,1,3);
        var0(:,:,:) = reshape([var(above) var(below) var(above)], 1, 1, 3);

        % The weight of B on O.
        % weight0 = zeros(symbols in O, symbls in B, states in H');
        weight0 = zeros(1,2,2)+0.01;

        % Observed nodes.
        bnet.CPD{i+3} = gaussian_CPD(bnet, bnet.rep_of_eclass(i+3),...
            'mean', mean0, 'cov', var0, 'weights', weight0, 'cov_prior_weight', 0);
    end

    engine = smoother_engine(jtree_2TBN_inf_engine(bnet));

    % Gather evidence.
    cases = cell(1,1);
    cases{1} = cell(ss, LEN);
    cases{1}(3,:) = num2cell([(xhyb > 0); ent], 1);
    cases{1}([4:ss],:) = num2cell(hyb);

    fprintf('Learning...\n');
    [bnet, LLtrace, engine] = learn_params_dbn_em(engine, cases);

    fprintf('Tabulating...\n');

    % Store the learned parameters.
    s = struct(bnet.CPD{ss+1});
    model.psnl = s.CPT(6, 18);
    model.pnl = s.CPT(7, 18);
    model.pfl = s.CPT(17, 18);
    model.pln = s.CPT(18, 1);
    model.pll = s.CPT(18, 18);
    model.mean_n = zeros(R, 1);
    model.mean_f = zeros(R, 1);
    model.mean_l = zeros(R, 1);
    model.var_n = zeros(R, 1);
    model.var_f = zeros(R, 1);
    model.var_l = zeros(R, 1);
    model.blast_weight_n = zeros(R, 1);
    model.blast_weight_f = zeros(R, 1);
    model.blast_weight_l = zeros(R, 1);
    model.entropy_weight_n = zeros(R, 1);
    model.entropy_weight_f = zeros(R, 1);
    model.entropy_weight_l = zeros(R, 1);
    for i = 1:R
        s = struct(bnet.CPD{i+3});
        model.mean_n(i) = s.mean(1,1);
        model.mean_f(i) = s.mean(1,2);
        model.mean_l(i) = s.mean(1,3);
        model.var_n(i) = s.cov(1,1,1);
        model.var_f(i) = s.cov(1,1,2);
        model.var_l(i) = s.cov(1,1,3);
        model.blast_weight_n(i) = s.weights(1,1,1);
        model.blast_weight_f(i) = s.weights(1,1,2);
        model.blast_weight_l(i) = s.weights(1,1,3);
        model.entropy_weight_n(i) = s.weights(1,2,1);
        model.entropy_weight_f(i) = s.weights(1,2,2);
        model.entropy_weight_l(i) = s.weights(1,2,3);
    end
     if tie_fuz,
        model.mean_f = model.mean_n;
        model.var_f = model.var_n;
        model.blast_weight_f = model.blast_weight_n;
        model.entropy_weight_f = model.entropy_weight_n;
    end
end

% Infer hidden states.
[engine, model.ll] = enter_evidence(engine, cases{1});
model.posts = zeros(18, LEN);
for t = 1:LEN
    marg = marginal_nodes(engine, 1, t);
    model.posts(:,t) = marg.T;
end
